// Copyright (c) 2023 Yuki Kimoto
// MIT License

#ifndef SPVM_BYTECODE_BUILDER_H
#define SPVM_BYTECODE_BUILDER_H

#include "spvm_typedecl.h"

void SPVM_BYTECODE_BUILDER_build_bytecode_array(SPVM_COMPILER* compiler);
void SPVM_OPCODE_BUILDER_build_opcode_array(SPVM_COMPILER* compiler);

#endif
